// Copyright 2024 Vincent Jacques

#ifndef LINCS__INTERNAL_HPP
#define LINCS__INTERNAL_HPP

namespace lincs {

// Mark parts of the API as "internal" where we cannot make them actually 'private'.
struct Internal {};

}  // namespace lincs

#endif  // LINCS__INTERNAL_HPP
